/* ######################################################################### */
/* ##                                                                     ## */
/* ##  S T A C K T H R U                                                  ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Job .......: Client part of a program to send data from host A     ## */
/* ##               B without processing them (with the goal to test for  ## */
/* ##               maximum thruput the stack allows).                    ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Copyright (C) 2002  Daniel Scheibli                                ## */
/* ##                                                                     ## */
/* ##  This program is free software; you can redistribute it and/or      ## */
/* ##  modify it under the terms of the GNU General Public License        ## */
/* ##  as published by the Free Software Foundation; either version 2     ## */
/* ##  of the License, or (at your option) any later version.             ## */
/* ##                                                                     ## */
/* ##  This program is distributed in the hope that it will be useful,    ## */
/* ##  but WITHOUT ANY WARRANTY; without even the implied warranty of     ## */
/* ##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      ## */
/* ##  GNU General Public License for more details.                       ## */
/* ##                                                                     ## */
/* ##  You should have received a copy of the GNU General Public License  ## */
/* ##  along with this program; if not, write to the Free Software        ## */
/* ##  Foundation, Inc., 59 Temple Place - Suite 330, Boston,             ## */
/* ##  MA  02111-1307, USA.                                               ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Notes .....: TCP  Header = 32 Bytes                                ## */
/* ##               UDP  Header =  8 Bytes                                ## */
/* ##               IPv4 Header = 20 Bytes                                ## */
/* ##                                                                     ## */
/* ## ------------------------------------------------------------------- ## */
/* ##                                                                     ## */
/* ##  Author ....: Daniel Scheibli <daniel@scheibli.com>                 ## */
/* ##  Date ......: 2002-03-05                                            ## */
/* ##  Changes ...: ----- stackthru V00.99.006 -------------------------- ## */
/* ##               2006-12-02 o Update code layout and program output.   ## */
/* ##               ----- stackthru V00.99.005 -------------------------- ## */
/* ##               2004-06-18 o Applied new code layout.                 ## */
/* ##               ----- stackthru V00.99.004 -------------------------- ## */
/* ##               2004-05-31 o Changed the sample value for block       ## */
/* ##                            size according to a MTU of 1500 bytes.   ## */
/* ##                          o Changed the Mbit value to be a power     ## */
/* ##                            of 10 instead of 2 (accroding to the     ## */
/* ##                            IEC standard).                           ## */
/* ##               2003-05-21 o Changed UdpV4 implementation.            ## */
/* ##               ----- stackthru V00.99.003 -------------------------- ## */
/* ##               2003-02-05 o Improved the output of the results.      ## */
/* ##               2003-02-05 o Added computation of the potential       ## */
/* ##                            number of frames sent.                   ## */
/* ##               ----- stackthru V00.99.002 -------------------------- ## */
/* ##               <none/>                                               ## */
/* ##               ----- stackthru V00.99.001 -------------------------- ## */
/* ##               <none/>                                               ## */
/* ##                                                                     ## */
/* ######################################################################### */



// Presettings

#include<stdlib.h>
#include<stdio.h>
#include<errno.h>
#include<math.h>
#include<time.h>
#include<strings.h>
#include<sys/timeb.h>
#include<netdb.h>
#include<sys/types.h>
#include<sys/socket.h>
#include<netinet/in.h>

#define TRUE      ( 1 == 1 )
#define FALSE     ( 1 == 2 )

#define PRINTF    printf
#define STRICMP   strcasecmp

#define EOS       '\0'
#define NEWLINE   "\n"



// Globale variables

char          *pcParamServerAddress;
char          *pcParamCommunication; 
unsigned long  ulParamBlockSize;
unsigned int   uiParamDataSize; 



// Function prototypes

int Communication_TcpV4( void );
int Communication_UdpV4( void );





/* ######################################################################### */
/* ##                                                                     ## */
/* ##   < M A I N >                                                       ## */
/* ##                                                                     ## */
/* ######################################################################### */
int main( int argc, char *argv[] ) {

	// Presettings

	int iRc;



	// Process parameters

	PRINTF( "%s", NEWLINE );
	PRINTF( "STACKTHRU V00.99.006%s", NEWLINE );
	PRINTF( "%s", NEWLINE );

	if( argc != ( 1 + 4 ) ) {
		PRINTF( "Usage ........: stackthru >>--- SERVERADDRESS --- COMMUNICATION --->  %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "                           >--- BLOCKSIZE --- DATASIZE ---><          %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "Parameters ...: SERVERADDRESS = Address where the stackthru-daemon    %s", NEWLINE );
		PRINTF( "                                is listening:                         %s", NEWLINE );
		PRINTF( "                                - For TCP/IP family: <host>:<port>    %s", NEWLINE );  
		PRINTF( "                COMMUNICATION = The protocol/way which should be used:%s", NEWLINE );
		PRINTF( "                                - TCPv4 = TCP over IP Version 4       %s", NEWLINE );  
		PRINTF( "                                - UDPv4 = UDP over IP Version 4       %s", NEWLINE );
		PRINTF( "                BLOCKSIZE     = Number of bytes per write() to the    %s", NEWLINE );
		PRINTF( "                                stackthru-daemon.                     %s", NEWLINE );
		PRINTF( "                DATASIZE      = Total size in MB of the data to be    %s", NEWLINE );
		PRINTF( "                                send to the stackthru-daemon.         %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "Sample .......: stackthru 192.168.2.1:2342 TCPv4 1448 256             %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		PRINTF( "                Sends 256 Megabytes of data (in blocks of 1448 Bytes; %s", NEWLINE );
		PRINTF( "                based on a MTU of 1500 Bytes) via TCP to port 2342    %s", NEWLINE );
		PRINTF( "                on the host 192.168.2.1 (where the stackthru daemon   %s", NEWLINE );
		PRINTF( "                is listening).                                        %s", NEWLINE );
		PRINTF( "%s", NEWLINE );
		exit( 1 );
	}
	else {
		pcParamServerAddress = argv[1];
		pcParamCommunication = argv[2];  
		ulParamBlockSize     = atol( argv[3] );
		uiParamDataSize      = atoi( argv[4] );
	}



	// Call the selected communication method

	iRc = -1;

	if( STRICMP( pcParamCommunication, "TCPV4" ) == 0 ) {
		iRc = Communication_TcpV4( );
	}
	if( STRICMP( pcParamCommunication, "UDPV4" ) == 0 ) {
		iRc = Communication_UdpV4( );
	}

	if( iRc != 0 ) {
		PRINTF( "===> ERROR: The selected testcase does not exists or returned%s", NEWLINE );
		PRINTF( "            with an errror (RC: %i)%s", iRc, NEWLINE );
		exit( 1 );
	}



	// Return program status

	return( 0 );
}





/* ######################################################################### */
/* ##                                                                     ## */
/* ##   C O M M U N I C A T I O N _ T C P V 4                             ## */
/* ##                                                                     ## */
/* ######################################################################### */
int Communication_TcpV4( void ) {

	// Presettings

	char                *pcParamHostToken = pcParamServerAddress;
	char                *pcParamPortToken = NULL; 
	int                  iParamPort       = -1;

	char                *pcFrameData      = NULL;
	unsigned long        ulIterations;
	double               ufIterations;
	unsigned long        ulLastIteration;

	struct hostent      *psHostInfo       = NULL;
	int                  iSocketFd;
	struct sockaddr_in   sInetAddr;

	struct in_addr     **ppsIpAddressForOutput;

	clock_t              xClockTickStart;
	clock_t              xClockTickStop; 
	struct timeb         sTimeStart;
	struct timeb         sTimeStop; 
	float                fTimeReal;

	int                  iRc;
	unsigned long        I;



	// Seperate host from port, convert and check port string

	for( I = 0; I < strlen( pcParamServerAddress ); I ++ ) {
		if( pcParamServerAddress[I] == ':' ) {
			pcParamServerAddress[I] = EOS;
			pcParamPortToken        = &pcParamServerAddress[(I + 1)];
			break;
		}
	} 
	if( pcParamPortToken != NULL ) {
		iParamPort = atoi( pcParamPortToken );
	}
	if( iParamPort <= 0 ) {
		PRINTF( "===> ERROR: The format of the given port is invalid!%s", NEWLINE );
		return( 1 );
	}


 
	// Build buffer for frame data to be send
	// (using random data and leading/trailing eyecatcher)

	pcFrameData = malloc( ulParamBlockSize );
	if( pcFrameData == NULL ) {
		PRINTF( "===> ERROR: Allocation of memory for frame data failed!%s", NEWLINE );
		return( 1 );
	}
	for( I = 0; I < ulParamBlockSize; I ++ ) {
		pcFrameData[I] = ( ( rand( ) + 1 ) % 224 ) + 32;
	}
	memcpy( &pcFrameData[0],                       "##### BEGIN #####", 17 );
	memcpy( &pcFrameData[(ulParamBlockSize - 15)], "##### END #####",   15 );



	// Calculate the number of sends/writes

	ufIterations    = ( ( ( 1.0 * uiParamDataSize ) / ulParamBlockSize ) * 1024 * 1024 );
	ulIterations    = floor( ufIterations );
	ulLastIteration = ( ufIterations - floor( ufIterations ) ) * ulParamBlockSize;



	// Convert hostname to IP address

	psHostInfo = gethostbyname( pcParamHostToken );
	if( psHostInfo == NULL ) {
		PRINTF( "===> ERROR: The given hostname could not be resolved!%s", NEWLINE );
		return( 1 );
	}



	// Before tasks

	ppsIpAddressForOutput = ( struct in_addr ** ) psHostInfo->h_addr_list;

	PRINTF( "Server.Name .............: %s%s", pcParamHostToken,                     NEWLINE );
	PRINTF( "Server.IP ...............: %s%s", inet_ntoa( **ppsIpAddressForOutput ), NEWLINE ); 
	PRINTF( "Server.Port .............: %i%s", iParamPort,                           NEWLINE );
	PRINTF( "%s", NEWLINE );
	PRINTF( "Transfer.Method .........: %s%s",         "TCP over IP Version 4",      NEWLINE );
	PRINTF( "Transfer.Size.Total .....: %u MB%s",      uiParamDataSize,              NEWLINE );
	PRINTF( "Transfer.Size.Block .....: %lu Bytes%s",  ulParamBlockSize,             NEWLINE ); 
	PRINTF( "Transfer.Blocks .........: %lu (%.3f)%s", ulIterations, ufIterations,   NEWLINE ); 
	PRINTF( "Transfer.Size.X-Block ...: %lu Bytes%s",  ulLastIteration,              NEWLINE );
	PRINTF( "%s", NEWLINE );

	iRc = ftime( &sTimeStart );
	xClockTickStart = clock( );



	// Create socket file descriptor

	iSocketFd = socket( AF_INET, SOCK_STREAM, 0 );
	if( iSocketFd < 0 ) {
		PRINTF( "===> ERROR: Failed while creating socket!%s", NEWLINE );
		return( 1 );
	}



	// Establish connection to the server

	sInetAddr.sin_family = AF_INET;
	sInetAddr.sin_port   = htons( iParamPort );
	memcpy( &sInetAddr.sin_addr, psHostInfo->h_addr, psHostInfo->h_length );

	iRc = connect( iSocketFd, ( struct sockaddr * ) &sInetAddr, sizeof( sInetAddr ) );
	if( iRc != 0 ) {
		PRINTF( "===> ERROR: Connection to server could not be established!%s", NEWLINE );
		return( 1 );
	}



	// Send the data

	for( I = 0; I < ulIterations; I ++ ) {
		if( write( iSocketFd, pcFrameData, ulParamBlockSize ) != ulParamBlockSize ) {
			PRINTF( "===> ERROR: Send of data (write) failed (1)!%s", NEWLINE );
			return( 1 );
		}
	}
	if( ulLastIteration > 0 ) {
		if( write( iSocketFd, pcFrameData, ulLastIteration ) != ulLastIteration ) {
			PRINTF( "===> ERROR: Send of data (write) failed (2)!%s", NEWLINE );
			return( 1 );
		}
	}



	// Close connection/socket
	
	close( iSocketFd );



	// After tasks

	xClockTickStop = clock( );
	iRc = ftime( &sTimeStop );

	fTimeReal = ( float ) ( ( sTimeStop.time - sTimeStart.time ) + ( ( sTimeStop.millitm - sTimeStart.millitm ) / 1000.0 ) );

	PRINTF( "Result.Time.Real ........: %.3f sec%s", fTimeReal, NEWLINE );
	PRINTF( "Result.Time.Sys .........: %.3f sec%s", ( float ) ( xClockTickStop - xClockTickStart ) / CLOCKS_PER_SEC,  NEWLINE );
	PRINTF( "%s", NEWLINE );
	PRINTF( "Result.Throughput .......: %.3f MB/sec (2^20)%s",   ( float ) ( 1.0 * uiParamDataSize / fTimeReal ),      NEWLINE );
	PRINTF( "Result.Throughput .......: %.3f Mbit/sec (10^6)%s", ( float ) ( 8.0 * ( 1024.0 / 1000 ) * ( 1024.0 / 1000 ) * uiParamDataSize / fTimeReal ), NEWLINE );
	PRINTF( "Result.Throughput .......: %.3f Frames/sec *%s",    ( float ) ( 1.0 * ( ulIterations + 1 ) / fTimeReal ), NEWLINE );  
	PRINTF( "%s", NEWLINE );
	PRINTF( "* = Only a rough estimate that can be far off - so handle with care! %s", NEWLINE );
	PRINTF( "    The calcuated value assumes that the data got transmitted with   %s", NEWLINE );
        PRINTF( "    TCP packets where the payload size equals the given block size.  %s". NEWLINE );
	PRINTF( "%s", NEWLINE );



	// End function

	return( 0 );
}





/* ######################################################################### */
/* ##                                                                     ## */
/* ##   C O M M U N I C A T I O N _ U D P V 4                             ## */
/* ##                                                                     ## */
/* ######################################################################### */
int Communication_UdpV4( void ) {

	// Presettings

	char                *pcParamHostToken = pcParamServerAddress;
	char                *pcParamPortToken = NULL; 
	int                  iParamPort       = -1;

	char                *pcFrameData      = NULL;
	unsigned long        ulIterations;
	double               ufIterations;
	unsigned long        ulLastIteration; 

	struct hostent      *psHostInfo       = NULL;
	int                  iSocketFd;
	struct sockaddr_in   sInetAddr;

	struct in_addr     **ppsIpAddressForOutput;

	clock_t              xClockTickStart;
	clock_t              xClockTickStop; 
	struct timeb         sTimeStart;
	struct timeb         sTimeStop; 
	float                fTimeReal;

	int                  iRc;
	unsigned long        I;



	// Seperate host from port, convert and check port string

	for( I = 0; I < strlen( pcParamServerAddress ); I ++ ) {
		if( pcParamServerAddress[I] == ':' ) {
			pcParamServerAddress[I] = EOS;
			pcParamPortToken        = &pcParamServerAddress[(I + 1)];
			break;
		}
	} 
	if( pcParamPortToken != NULL ) {
		iParamPort = atoi( pcParamPortToken );
	}
	if( iParamPort <= 0 ) {
		PRINTF( "===> ERROR: The format of the given port is invalid!%s", NEWLINE );
		return( 1 );
	}


 
	// Build buffer for frame data to be send
	// (using nearly uncompressable data and eyecatchers)

	pcFrameData = malloc( ulParamBlockSize );
	if( pcFrameData == NULL ) {
		PRINTF( "===> ERROR: Allocation of memory for frame data failed!%s", NEWLINE );
		return( 1 );
	}
	for( I = 0; I < ulParamBlockSize; I ++ ) {
		pcFrameData[I] = ( ( rand( ) + 1 ) % 224 ) + 32;
	}
	memcpy( &pcFrameData[0],                       "##### BEGIN #####", 17 );
	memcpy( &pcFrameData[(ulParamBlockSize - 15)], "##### END #####",   15 );



	// Calculate the number of sends/writes

	ufIterations    = ( ( ( 1.0 * uiParamDataSize ) / ulParamBlockSize ) * 1024 * 1024 );
	ulIterations    = floor( ufIterations );
	ulLastIteration = ( ufIterations - floor( ufIterations ) ) * ulParamBlockSize;



	// Convert hostname to IP address

	psHostInfo = gethostbyname( pcParamHostToken );
	if( psHostInfo == NULL ) {
		PRINTF( "===> ERROR: The given hostname could not be resolved!%s", NEWLINE );
		return( 1 );
	}
	sInetAddr.sin_family = AF_INET;
	sInetAddr.sin_port   = htons( iParamPort );
	memcpy( &sInetAddr.sin_addr, psHostInfo->h_addr, psHostInfo->h_length );



	// Before tasks

	ppsIpAddressForOutput = ( struct in_addr ** ) psHostInfo->h_addr_list;

	PRINTF( "Server.Name .............: %s%s", pcParamHostToken,                     NEWLINE );
	PRINTF( "Server.IP ...............: %s%s", inet_ntoa( **ppsIpAddressForOutput ), NEWLINE ); 
	PRINTF( "Server.Port .............: %i%s", iParamPort,                           NEWLINE );
	PRINTF( "%s", NEWLINE );
	PRINTF( "Transfer.Method .........: %s%s",         "UDP over IP Version 4",      NEWLINE );
	PRINTF( "Transfer.Size.Total .....: %u MB%s",      uiParamDataSize,              NEWLINE );
	PRINTF( "Transfer.Size.Block .....: %lu Bytes%s",  ulParamBlockSize,             NEWLINE ); 
	PRINTF( "Transfer.Blocks .........: %lu (%.3f)%s", ulIterations, ufIterations,   NEWLINE ); 
	PRINTF( "Transfer.Size.X-Block ...: %lu Bytes%s",  ulLastIteration,              NEWLINE );  
	PRINTF( "%s", NEWLINE );

	iRc = ftime( &sTimeStart );
	xClockTickStart = clock( );



	// Create socket file descriptor

 	iSocketFd = socket( AF_INET, SOCK_DGRAM, 0 );
	if( iSocketFd < 0 ) {
		PRINTF( "===> ERROR: Failed while creating socket!%s", NEWLINE );
		return( 1 );
	}



	// Send the data

	for( I = 0; I < ulIterations; I ++ ) {
		if( sendto( iSocketFd, pcFrameData, ulParamBlockSize, 0, ( struct sockaddr * ) &sInetAddr, sizeof( sInetAddr ) ) != ulParamBlockSize ) {
			PRINTF( "===> ERROR: Send of data (write) failed (1)!%s", NEWLINE );
			return( 1 );   
		}
	}
	if( ulLastIteration > 0 ) {
		if( sendto( iSocketFd, pcFrameData, ulLastIteration, 0, ( struct sockaddr * ) &sInetAddr, sizeof( sInetAddr ) ) != ulLastIteration ) {
			PRINTF( "===> ERROR: Send of data (write) failed (2)!%s", NEWLINE );
			return( 1 );   
		}
	}



	// After tasks

	xClockTickStop = clock( );
	iRc = ftime( &sTimeStop );

	fTimeReal = ( float ) ( ( sTimeStop.time - sTimeStart.time ) + ( ( sTimeStop.millitm - sTimeStart.millitm ) / 1000.0 ) );

	PRINTF( "Result.Time.Real ........: %.3f sec%s", fTimeReal, NEWLINE );
	PRINTF( "Result.Time.Sys .........: %.3f sec%s", ( float ) ( xClockTickStop - xClockTickStart ) / CLOCKS_PER_SEC, NEWLINE );
	PRINTF( "%s", NEWLINE );
	PRINTF( "Result.Throughput .......: %.3f MB/sec%s",       ( float ) ( 1.0 * uiParamDataSize / fTimeReal ),      NEWLINE );
	PRINTF( "Result.Throughput .......: %.3f Mbit/sec%s",     ( float ) ( 8.0 * uiParamDataSize / fTimeReal ),      NEWLINE ); 
	PRINTF( "Result.Throughput .......: %.3f Frames/sec *%s", ( float ) ( 1.0 * ( ulIterations + 1 ) / fTimeReal ), NEWLINE );  
	PRINTF( "%s", NEWLINE );
	PRINTF( "* = Only valid, if the used block size is less or equal to the maximum size%s", NEWLINE );
	PRINTF( "    of a frames data part (e.g. 1472 Bytes for UDP over IP version 4; in an%s", NEWLINE );  
	PRINTF( "    environment with a MTU size of 1500 Bytes).%s",                             NEWLINE );  
	PRINTF( "%s", NEWLINE );



	// Close connection/socket

	close( iSocketFd );



	// End function

	return( 0 );
}




